@@ -1,5 +1,5 @@
 /*
- * Copyright 2008-2018 Aerospike, Inc.
+ * Copyright 2008-2025 Aerospike, Inc.
  *
  * Portions may be licensed to Aerospike, Inc. under one or more contributor
  * license agreements.
@@ -26,36 +26,22 @@
 extern "C" {
 #endif
 
-/******************************************************************************
- * MACROS
- *****************************************************************************/
+//---------------------------------
+// Macros
+//---------------------------------
 
 #ifdef __linux__
-/**
- * Default path to the system UDF files.
- */
-#define AS_CONFIG_LUA_SYSTEM_PATH "/opt/aerospike/client/sys/udf/lua"
-
 /**
  * Default path to the user UDF files.
  */
 #define AS_CONFIG_LUA_USER_PATH "/opt/aerospike/client/usr/udf/lua"
 #endif
 
-#ifdef __APPLE__
-/**
- * Default path to the system UDF files.
- */
-#define AS_CONFIG_LUA_SYSTEM_PATH "/usr/local/aerospike/client/sys/udf/lua"
-
-/**
- * Default path to the user UDF files.
- */
+#if defined(__APPLE__) || defined(__FreeBSD__)
 #define AS_CONFIG_LUA_USER_PATH "/usr/local/aerospike/client/usr/udf/lua"
 #endif
 
 #if defined(_MSC_VER)
-#define AS_CONFIG_LUA_SYSTEM_PATH "C:/aerospike/client/sys/udf/lua"
 #define AS_CONFIG_LUA_USER_PATH "C:/aerospike/client/usr/udf/lua"
 #endif
 
@@ -74,14 +60,24 @@ extern "C" {
  */
 #define AS_PASSWORD_SIZE 64
 
-/******************************************************************************
- * TYPES
- *****************************************************************************/
+/**
+ * Minimum tend interval in milliseconds.
+ */
+#define AS_TEND_INTERVAL_MIN 250
+
+/**
+ * Default interval in milliseconds between dynamic configuration check for file modifications.
+ */
+#define AS_CONFIG_PROVIDER_INTERVAL_DEFAULT 60000
+
+//---------------------------------
+// Types
+//---------------------------------
 
 /**
  * IP translation table.
  *
- * @ingroup as_config_object
+ * @relates as_config
  */
 typedef struct as_addr_map_s {
 	
@@ -97,10 +93,44 @@ typedef struct as_addr_map_s {
 	
 } as_addr_map;
 
+/**
+ * Authentication mode.
+ *
+ * @relates as_config
+ */
+typedef enum as_auth_mode_e {
+	/**
+	 * Use internal authentication when user/password defined. Hashed password is stored
+	 * on the server. Do not send clear password. This is the default.
+	 */
+	AS_AUTH_INTERNAL,
+
+	/**
+	 * Use external authentication (like LDAP) when user/password defined. Specific external
+	 * authentication is configured on server. If TLS defined, send clear password on node
+	 * login via TLS. Return error if TLS is not defined.
+	 */
+	AS_AUTH_EXTERNAL,
+
+	/**
+	 * Use external authentication (like LDAP) when user/password defined. Specific external
+	 * authentication is configured on server.  Send clear password on node login whether or
+	 * not TLS is defined. This mode should only be used for testing purposes because it is
+	 * not secure authentication.
+	 */
+	AS_AUTH_EXTERNAL_INSECURE,
+
+	/**
+	 * Authentication and authorization based on a certificate.  No user name or
+	 * password needs to be configured.  Requires TLS and a client certificate.
+	 */
+	AS_AUTH_PKI
+} as_auth_mode;
+
 /**
  * Cluster event notification type.
  *
- * @ingroup as_config_object
+ * @relates as_config
  */
 typedef enum as_cluster_event_type_e {
 	/**
@@ -122,7 +152,7 @@ typedef enum as_cluster_event_type_e {
 /**
  * Cluster event notification data.
  *
- * @ingroup as_config_object
+ * @relates as_config
  */
 typedef struct as_cluster_event_s {
 	/**
@@ -151,14 +181,14 @@ typedef struct as_cluster_event_s {
  * as_cluster_event is placed on the stack before calling.
  * Do not free node_name or node_address.
  *
- * @ingroup as_config_object
+ * @relates as_config
  */
 typedef void (*as_cluster_event_callback) (as_cluster_event* event);
 
 /**
  * lua module config
  *
- * @ingroup as_config_object
+ * @relates as_config
  */
 typedef struct as_config_lua_s {
 
@@ -168,13 +198,6 @@ typedef struct as_config_lua_s {
 	 */
 	bool cache_enabled;
 
-	/**
-	 * The path to the system UDF files. These UDF files 
-	 * are installed with the aerospike client library.
-	 * Default location defined in: AS_CONFIG_LUA_SYSTEM_PATH
-	 */
-	char system_path[AS_CONFIG_PATH_MAX_SIZE];
-
 	/**
 	 * The path to user's UDF files.
 	 * Default location defined in: AS_CONFIG_LUA_USER_PATH
@@ -186,7 +209,7 @@ typedef struct as_config_lua_s {
 /**
  * TLS module config
  *
- * @ingroup as_config_object
+ * @relates as_config
  */
 typedef struct as_config_tls_s {
 
@@ -200,9 +223,17 @@ typedef struct as_config_tls_s {
 	 * Path to a trusted CA certificate file.
 	 * By default TLS will use system standard trusted CA certificates.
 	 * Use as_config_tls_set_cafile() to set this field.
+	 * If cafile is populated, castring is ignored.
 	 */
 	char* cafile;
 
+	/**
+	 * String containing trusted CA certificate(s).
+	 * Use as_config_tls_set_castring() to set this field.
+	 * If cafile is populated, castring is ignored.
+	 */
+	char* castring;
+
 	/**
 	 * Path to a directory of trusted certificates.
 	 * See the OpenSSL SSL_CTX_load_verify_locations manual page for
@@ -242,20 +273,6 @@ typedef struct as_config_tls_s {
 	 */
 	char* cipher_suite;
 	
-	/**
-	 * Enable CRL checking for the certificate chain leaf certificate.
-	 * An error occurs if a suitable CRL cannot be found.
-     * By default CRL checking is disabled.
-	 */
-	bool crl_check;
-
-	/**
-	 * Enable CRL checking for the entire certificate chain.
-	 * An error occurs if a suitable CRL cannot be found.
-     * By default CRL checking is disabled.
-	 */
-	bool crl_check_all;
-
 	/**
 	 * Path to a certificate blacklist file.
 	 * The file should contain one line for each blacklisted certificate.
@@ -270,29 +287,101 @@ typedef struct as_config_tls_s {
 	 */
 	char* cert_blacklist;
 
-	/**
-	 * Log session information for each connection.
-	 */
-	bool log_session_info;
-	
 	/**
 	 * Path to the client's key for mutual authentication.
 	 * By default mutual authentication is disabled.
-	 *
 	 * Use as_config_tls_set_keyfile() to set this field.
+	 * If keyfile is populated, keystring is ignored.
 	 */
 	char* keyfile;
 
+	/**
+	 * Decryption password for the client's key for mutual authentication.
+	 * By default the key is assumed not to be encrypted.
+	 *
+	 * Use as_config_tls_set_keyfile_pw() to set this field.
+	 */
+	char* keyfile_pw;
+
+	/**
+	 * Client's key string for mutual authentication.
+	 * By default mutual authentication is disabled.
+	 * Use as_config_tls_set_keystring() to set this field.
+	 * If keyfile is populated, keystring is ignored.
+	 */
+	char* keystring;
+
 	/**
 	 * Path to the client's certificate chain file for mutual authentication.
 	 * By default mutual authentication is disabled.
-	 *
 	 * Use as_config_tls_set_certfile() to set this field.
+	 * If certfile is populated, certstring is ignored.
 	 */
 	char* certfile;
 
+	/**
+	 * Client's certificate chain file string for mutual authentication.
+	 * By default mutual authentication is disabled.
+	 * Use as_config_tls_set_certstring() to set this field.
+	 * If certfile is populated, certstring is ignored.
+	 */
+	char* certstring;
+
+	/**
+	 * Enable CRL checking for the certificate chain leaf certificate.
+	 * An error occurs if a suitable CRL cannot be found.
+	 * By default CRL checking is disabled.
+	 */
+	bool crl_check;
+
+	/**
+	 * Enable CRL checking for the entire certificate chain.
+	 * An error occurs if a suitable CRL cannot be found.
+	 * By default CRL checking is disabled.
+	 */
+	bool crl_check_all;
+
+	/**
+	 * Log session information for each connection.
+	 */
+	bool log_session_info;
+
+	/**
+	 * Use TLS connections only for login authentication.
+	 * All other communication with the server will be done with non-TLS connections.
+	 * Default: false (Use TLS connections for all communication with server.)
+	 */
+	bool for_login_only;
+
 } as_config_tls;
 
+/**
+ * Dynamic configuration provider. Determines how to retrieve cluster policies.
+ *
+ * @relates as_config
+ */
+typedef struct as_config_provider_s {
+
+	/**
+	 * Dynamic configuration file path. If set, cluster policies will be read from the yaml file at cluster
+	 * initialization and whenever the file changes. The policies fields in the file
+	 * override all command policies.
+	 *
+	 * Use as_config_provider_set_path() to set this field.
+	 * Default: NULL
+	 */
+	char* path;
+
+	/**
+	 * Interval in milliseconds between dynamic configuration check for file modifications.
+	 * The value must be greater than or equal to the tend interval.
+	 *
+	 * Default: 60000
+	 */
+	uint32_t interval;
+
+} as_config_provider;
+
 /**
  * The `as_config` contains the settings for the `aerospike` client. Including
  * default policies, seed hosts in the cluster and other settings.
@@ -302,10 +391,10 @@ typedef struct as_config_tls_s {
  * Before using as_config, you must first initialize it. This will setup the 
  * default values.
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * as_config config;
  * as_config_init(&config);
- * ~~~~~~~~~~
+ * @endcode
  *
  * Once initialized, you can populate the values.
  *
@@ -314,9 +403,9 @@ typedef struct as_config_tls_s {
  * The client will require at least one seed host defined in the 
  * configuration. The seed host is defined in `as_config.hosts`. 
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * as_config_add_host(&config, "127.0.0.1", 3000);
- * ~~~~~~~~~~
+ * @endcode
  *
  * The client will iterate over the list until it connects with one of the hosts.
  *
@@ -345,9 +434,9 @@ typedef struct as_config_tls_s {
  * to `aerospike_key_put()`, then you may find it beneficial to set the global
  * `as_policy_write` in `as_policies.write`, which all write operations will use.
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * config.policies.write.key = AS_POLICY_KEY_SEND;
- * ~~~~~~~~~~
+ * @endcode
  *
  * Global operation policies:
  * - as_policies.read
@@ -366,12 +455,9 @@ typedef struct as_config_tls_s {
  * which allows you to define a path to where the client library will look for
  * Lua files for processing.
  * 
- * ~~~~~~~~~~{.c}
+ * @code
  * strcpy(config.mod_lua.user_path, "/home/me/lua");
- * ~~~~~~~~~~
- *
- * Never call as_config_destroy() directly because ownership of config fields
- * is transferred to aerospike in aerospike_init() or aerospike_new().
+ * @endcode
  *
  * @ingroup client_objects
  */
@@ -393,13 +479,24 @@ typedef struct as_config_s {
 	char password[AS_PASSWORD_SIZE];
 	
 	/**
-	 * Expected cluster name.  If not null, server nodes must return this cluster name in order to
-	 * join the client's view of the cluster. Should only be set when connecting to servers that
-	 * support the "cluster-name" info command.  Use as_config_set_cluster_name() to set this field.
+	 * Expected cluster name.  If populated, the cluster_name must match the cluster-name field
+	 * in the service section in each server configuration.  This ensures that the specified
+	 * seed nodes belong to the expected cluster on startup.  If not, the client will refuse
+	 * to add the node to the client's view of the cluster.
+	 *
+	 * Use as_config_set_cluster_name() to set this field.
+	 *
 	 * Default: NULL
 	 */
 	char* cluster_name;
 	
+	/**
+	 * Application identifier. Use as_config_set_app_id() to set this field.
+	 *
+	 * Default: NULL
+	 */
+	char* app_id;
+
 	/**
 	 * Cluster event function that will be called when nodes are added/removed from the cluster.
 	 *
@@ -434,25 +531,58 @@ typedef struct as_config_s {
 	uint32_t ip_map_size;
 	
 	/**
-	 * Maximum number of synchronous connections allowed per server node.  Synchronous transactions
-	 * will go through retry logic and potentially fail with error code "AEROSPIKE_ERR_NO_MORE_CONNECTIONS"
-	 * if the maximum number of connections would be exceeded.
+	 * Minimum number of synchronous connections allowed per server node.  Preallocate min
+	 * connections on client node creation.  The client will periodically allocate new connections
+	 * if count falls below min connections.
+	 *
+	 * Server proto-fd-idle-ms and client max_socket_idle should be set to zero (no reap) if
+	 * min_conns_per_node is greater than zero.  Reaping connections can defeat the purpose
+	 * of keeping connections in reserve for a future burst of activity.
+	 *
+	 * Servers 8.1+ have deprecated proto-fd-idle-ms. When proto-fd-idle-ms is ultimately removed,
+	 * the server will stop automatically reaping based on socket idle timeouts.
+	 *
+	 * Default: 0
+	 */
+	uint32_t min_conns_per_node;
+
+	/**
+	 * Maximum number of synchronous connections allowed per server node.  Synchronous commands
+	 * will go through retry logic and potentially fail with error code
+	 * "AEROSPIKE_ERR_NO_MORE_CONNECTIONS" if the maximum number of connections would be exceeded.
 	 * 
 	 * The number of connections used per node depends on how many concurrent threads issue
 	 * database commands plus sub-threads used for parallel multi-node commands (batch, scan,
 	 * and query). One connection will be used for each thread.
 	 *
-	 * Default: 300
+	 * Default: 100
 	 */
 	uint32_t max_conns_per_node;
 	
+	/**
+	 * Minimum number of asynchronous connections allowed per server node.  Preallocate min
+	 * connections on client node creation.  The client will periodically allocate new connections
+	 * if count falls below min connections.
+	 *
+	 * Server proto-fd-idle-ms and client max_socket_idle should be set to zero (no reap) if
+	 * async_min_conns_per_node is greater than zero.  Reaping connections can defeat the purpose
+	 * of keeping connections in reserve for a future burst of activity.
+	 *
+	 * Servers 8.1+ have deprecated proto-fd-idle-ms. When proto-fd-idle-ms is ultimately removed,
+	 * the server will stop automatically reaping based on socket idle timeouts.
+	 *
+	 * Default: 0
+	 */
+	uint32_t async_min_conns_per_node;
+
 	/**
 	 * Maximum number of asynchronous (non-pipeline) connections allowed for each node.
 	 * This limit will be enforced at the node/event loop level.  If the value is 100 and 2 event
 	 * loops are created, then each node/event loop asynchronous (non-pipeline) connection pool 
-	 * will have a limit of 50. Async transactions will be rejected if the limit would be exceeded.
+	 * will have a limit of 50. Async commands will be rejected if the limit would be exceeded.
 	 * This variable is ignored if asynchronous event loops are not created.
-	 * Default: 300
+	 *
+	 * Default: 100
 	 */
 	uint32_t async_max_conns_per_node;
 
@@ -460,8 +590,9 @@ typedef struct as_config_s {
 	 * Maximum number of pipeline connections allowed for each node.
 	 * This limit will be enforced at the node/event loop level.  If the value is 100 and 2 event
 	 * loops are created, then each node/event loop pipeline connection pool will have a limit of 50. 
-	 * Async transactions will be rejected if the limit would be exceeded.
+	 * Async commands will be rejected if the limit would be exceeded.
 	 * This variable is ignored if asynchronous event loops are not created.
+	 *
 	 * Default: 64
 	 */
 	uint32_t pipe_max_conns_per_node;
@@ -478,24 +609,68 @@ typedef struct as_config_s {
 	uint32_t conn_pools_per_node;
 
 	/**
-	 * Initial host connection timeout in milliseconds.  The timeout when opening a connection
-	 * to the server host for the first time.
+	 * Cluster tend info command timeout in milliseconds.
+	 *
 	 * Default: 1000
 	 */
 	uint32_t conn_timeout_ms;
 
 	/**
-	 * Maximum socket idle time in seconds.  Connection pools will discard sockets that have
-	 * been idle longer than the maximum.  The value is limited to 24 hours (86400).
+	 * Node login timeout in milliseconds.
+	 * Default: 5000
+	 */
+	uint32_t login_timeout_ms;
+
+	/**
+	 * Maximum socket idle in seconds.  Connection pools will discard sockets that have been 
+	 * idle longer than the maximum.
 	 *
-	 * It's important to set this value to a few seconds less than the server's proto-fd-idle-ms
-	 * (default 60000 milliseconds or 1 minute), so the client does not attempt to use a socket
-	 * that has already been reaped by the server.
+	 * Connection pools are now implemented by a LIFO stack.  Connections at the tail of the
+	 * stack will always be the least used.  These connections are checked for max_socket_idle
+	 * once every 30 tend iterations (usually 30 seconds).
 	 *
-	 * Default: 0 seconds (disabled) for non-TLS connections, 55 seconds for TLS connections.
+	 * If server's proto-fd-idle-ms is greater than zero, then max_socket_idle should be
+	 * at least a few seconds less than the server's proto-fd-idle-ms, so the client does not
+	 * attempt to use a socket that has already been reaped by the server.
+	 *
+	 * If server's proto-fd-idle-ms is zero (no reap), then max_socket_idle should also be zero.
+	 * Connections retrieved from a pool in commands will not be checked for max_socket_idle
+	 * when max_socket_idle is zero.  Idle connections will still be trimmed down from peak
+	 * connections to min connections (min_conns_per_node and async_min_conns_per_node) using a
+	 * hard-coded 55 second limit in the cluster tend thread.
+	 *
+	 * Servers 8.1+ have deprecated proto-fd-idle-ms. When proto-fd-idle-ms is ultimately removed,
+	 * the server will stop automatically reaping based on socket idle timeouts.
+	 *
+	 * Default: 0 seconds
 	 */
 	uint32_t max_socket_idle;
 
+	/**
+	 * Maximum number of errors allowed per node per error_rate_window before backoff
+	 * algorithm returns AEROSPIKE_MAX_ERROR_RATE for database commands to that node.
+	 * If max_error_rate is zero, there is no error limit.
+	 *
+	 * The counted error types are any error that causes the connection to close (socket errors
+	 * and client timeouts), server device overload and server timeouts.
+	 *
+	 * The application should backoff or reduce the command load until AEROSPIKE_MAX_ERROR_RATE
+	 * stops being returned.
+	 *
+	 * Default: 100
+	 */
+	uint32_t max_error_rate;
+
+	/**
+	 * The number of cluster tend iterations that defines the window for max_error_rate.
+	 * One tend iteration is defined as tender_interval plus the time to tend all nodes.
+	 * At the end of the window, the error count is reset to zero and backoff state is removed
+	 * on all nodes.
+	 *
+	 * Default: 1
+	 */
+	uint32_t error_rate_window;
+
 	/**
 	 * Polling interval in milliseconds for cluster tender
 	 * Default: 1000
@@ -514,12 +689,23 @@ typedef struct as_config_s {
 	 * Default: 16
 	 */
 	uint32_t thread_pool_size;
-	
+
+	/**
+	 * Assign tend thread to this specific CPU ID.
+	 * Default: -1 (Any CPU).
+	 */
+	int tend_thread_cpu;
+
 	/**
 	 * Client policies
 	 */
 	as_policies policies;
 
+	/**
+	 * Dynamic configuration provider.
+	 */
+	as_config_provider config_provider;
+
 	/**
 	 * lua config.  This is a global config even though it's located here in cluster config.
 	 * This config has been left here to avoid breaking the API.
@@ -528,44 +714,108 @@ typedef struct as_config_s {
 	 * A better method for initializing lua configuration is to leave this field alone and
 	 * instead call aerospike_init_lua():
 	 *
-	 * ~~~~~~~~~~{.c}
+	 * @code
 	 * // Get default global lua configuration.
 	 * as_config_lua lua;
 	 * as_config_lua_init(&lua);
 	 *
 	 * // Optionally modify lua defaults.
 	 * lua.cache_enabled = <enable lua cache>;
-	 * strcpy(lua.system_path, <lua system directory>);
 	 * strcpy(lua.user_path, <lua user directory>);
 	 *
 	 * // Initialize global lua configuration.
 	 * aerospike_init_lua(&lua);
-	 * ~~~~~~~~~~
+	 * @endcode
 	 */
 	as_config_lua lua;
 
-	/*
+	/**
 	 * TLS configuration parameters.
 	 */
 	as_config_tls tls;
-	
+
 	/**
-	 * Action to perform if client fails to connect to seed hosts.
+	 * Authentication mode.
+	 * Default: AS_AUTH_INTERNAL
+	 */
+	as_auth_mode auth_mode;
+
+	/**
+	 * Should cluster instantiation fail if the client fails to connect to a seed or
+	 * all the seed's peers.
 	 *
-	 * If fail_if_not_connected is true (default), the cluster creation will fail
-	 * when all seed hosts are not reachable.
+	 * If true, return an error if all seed connections fail or a seed is valid,
+	 * but all peers from that seed are not reachable.
 	 *
-	 * If fail_if_not_connected is false, an empty cluster will be created and the 
-	 * client will automatically connect when Aerospike server becomes available.
+	 * If false, a partial cluster will be created and the client will automatically connect
+	 * to the remaining nodes when they become available.
 	 */
 	bool fail_if_not_connected;
 	
 	/**
-	 * Flag to signify if "services-alternate" should be used instead of "services"
-	 * Default : false
+	 * Flag to signify if alternate IP address discovery info commands should be used.
+	 *
+	 * If false, use:
+	 * IP address: service-clear-std
+	 * TLS IP address: service-tls-std
+	 * Peers addresses: peers-clear-std
+	 * Peers TLS addresses: peers-tls-std
+	 *
+	 * If true, use:
+	 * IP address: service-clear-alt
+	 * TLS IP address: service-tls-alt
+	 * Peers addresses: peers-clear-alt
+	 * Peers TLS addresses: peers-tls-alt
+	 *
+	 * Default: false
 	 */
 	bool use_services_alternate;
 
+	/**
+	 * For testing purposes only.  Do not modify.
+	 *
+	 * Should the aerospike instance communicate with the first seed node only
+	 * instead of using the data partition map to determine which node to send the
+	 * database command.
+	 *
+	 * Default: false
+	 */
+	bool force_single_node;
+
+	/**
+	 * Track server rack data.  This field is useful when directing read commands to 
+	 * the server node that contains the key and exists on the same rack as the client.
+	 * This serves to lower cloud provider costs when nodes are distributed across different
+	 * racks/data centers.
+	 *
+	 * rack_id or rack_ids, AS_POLICY_REPLICA_PREFER_RACK and server rack configuration must
+	 * also be set to enable this functionality.
+	 *
+	 * Default: false
+	 */
+	bool rack_aware;
+
+	/**
+	 * Rack where this client instance resides. If rack_ids is set, rack_id is ignored.
+	 *
+	 * rack_aware, AS_POLICY_REPLICA_PREFER_RACK and server rack configuration must also be
+	 * set to enable this functionality.
+	 *
+	 * Default: 0
+	 */
+	int rack_id;
+
+	/**
+	 * List of preferred racks in order of preference. If rack_ids is set, rack_id is ignored.
+	 * Do not set directly. Use multiple as_config_add_rack_id() calls to add rack ids.
+	 *
+	 * rack_aware, AS_POLICY_REPLICA_PREFER_RACK and server rack configuration must also be
+	 * set to enable this functionality.
+	 *
+	 * Default: NULL
+	 */
+	 as_vector* rack_ids;
+
 	/**
 	 * Indicates if shared memory should be used for cluster tending.  Shared memory
 	 * is useful when operating in single threaded mode with multiple client processes.
@@ -583,7 +833,7 @@ typedef struct as_config_s {
 	 * Each shared memory segment contains state for one Aerospike cluster.  If there are
 	 * multiple Aerospike clusters, a different shm_key must be defined for each cluster.
 	 * 
-	 * Default: 0xA7000000
+	 * Default: 0xA9000000
 	 */
 	int shm_key;
 	
@@ -611,9 +861,9 @@ typedef struct as_config_s {
 	uint32_t shm_takeover_threshold_sec;
 } as_config;
 
-/******************************************************************************
- * FUNCTIONS
- *****************************************************************************/
+//---------------------------------
+// Functions
+//---------------------------------
 
 /**
  * Initialize the configuration to default values.
@@ -621,35 +871,46 @@ typedef struct as_config_s {
  * You should do this to ensure the configuration has valid values, before 
  * populating it with custom options.
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * as_config config;
  * as_config_init(&config);
  * as_config_add_host(&config, "127.0.0.1", 3000);
- * ~~~~~~~~~~
+ * @endcode
  * 
  * @relates as_config
  */
 AS_EXTERN as_config*
 as_config_init(as_config* config);
 
+/**
+ * Destroy configuration.
+ *
+ * Do not call as_config_destroy() after calling aerospike_init() or aerospike_new() because
+ * ownership of config fields is transferred to the aerospike instance.
+ *
+ * @relates as_config
+ */
+AS_EXTERN void
+as_config_destroy(as_config* config);
+
 /**
  * Add seed host(s) from a string with format: hostname1[:tlsname1][:port1],...
  * Hostname may also be an IP address in the following formats.
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * IPv4: xxx.xxx.xxx.xxx
  * IPv6: [xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx]
  * IPv6: [xxxx::xxxx]
- * ~~~~~~~~~~
+ * @endcode
  *
  * The host addresses will be copied.
  * The caller is responsible for the original string.
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * as_config config;
  * as_config_init(&config);
  * as_config_add_hosts(&config, "host1,host2:3010,192.168.20.1:3020,[2001::1000]:3030", 3000);
- * ~~~~~~~~~~
+ * @endcode
  *
  * @relates as_config
  */
@@ -661,11 +922,11 @@ as_config_add_hosts(as_config* config, const char* string, uint16_t default_port
  * The host address will be copied.
  * The caller is responsible for the original address string.
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * as_config config;
  * as_config_init(&config);
  * as_config_add_host(&config, "127.0.0.1", 3000);
- * ~~~~~~~~~~
+ * @endcode
  *
  * @relates as_config
  */
@@ -684,11 +945,11 @@ as_config_clear_hosts(as_config* config);
  * User authentication for servers with restricted access.  The password will be stored by the
  * client and sent to server in hashed format.
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * 	as_config config;
  * 	as_config_init(&config);
  * 	as_config_set_user(&config, "charlie", "mypassword");
- * ~~~~~~~~~~
+ * @endcode
  *
  * @relates as_config
  */
@@ -712,6 +973,17 @@ as_config_set_cluster_name(as_config* config, const char* cluster_name)
 	as_config_set_string(&config->cluster_name, cluster_name);
 }
 
+/**
+ * Set application identifier.
+ *
+ * @relates as_config
+ */
+static inline void
+as_config_set_app_id(as_config* config, const char* app_id)
+{
+	as_config_set_string(&config->app_id, app_id);
+}
+
 /**
  * Set cluster event callback and user data.
  *
@@ -733,7 +1005,6 @@ static inline void
 as_config_lua_init(as_config_lua* lua)
 {
 	lua->cache_enabled = false;
-	strcpy(lua->system_path, AS_CONFIG_LUA_SYSTEM_PATH);
 	strcpy(lua->user_path, AS_CONFIG_LUA_USER_PATH);
 }
 
@@ -748,6 +1019,17 @@ as_config_tls_set_cafile(as_config* config, const char* cafile)
 	as_config_set_string(&config->tls.cafile, cafile);
 }
 
+/**
+ * Set string containing trusted CA certificate(s).
+ *
+ * @relates as_config
+ */
+static inline void
+as_config_tls_set_castring(as_config* config, const char* castring)
+{
+	as_config_set_string(&config->tls.castring, castring);
+}
+
 /**
  * Set TLS path to a directory of trusted certificates.
  *
@@ -803,6 +1085,28 @@ as_config_tls_set_keyfile(as_config* config, const char* keyfile)
 	as_config_set_string(&config->tls.keyfile, keyfile);
 }
 
+/**
+ * Set decryption password for the client's key.
+ *
+ * @relates as_config
+ */
+static inline void
+as_config_tls_set_keyfile_pw(as_config* config, const char* pw)
+{
+	as_config_set_string(&config->tls.keyfile_pw, pw);
+}
+
+/**
+ * Set client's key string for mutual authentication.
+ *
+ * @relates as_config
+ */
+static inline void
+as_config_tls_set_keystring(as_config* config, const char* keystring)
+{
+	as_config_set_string(&config->tls.keystring, keystring);
+}
+
 /**
  * Set TLS path to the client's certificate chain file for mutual authentication.
  *
@@ -814,22 +1118,98 @@ as_config_tls_set_certfile(as_config* config, const char* certfile)
 	as_config_set_string(&config->tls.certfile, certfile);
 }
 
+/**
+ * Set client's certificate chain file string for mutual authentication.
+ *
+ * @relates as_config
+ */
+static inline void
+as_config_tls_set_certstring(as_config* config, const char* certstring)
+{
+	as_config_set_string(&config->tls.certstring, certstring);
+}
+
 /**
  * Add TLS host to seed the cluster.
  * The host address and TLS name will be copied.
  * The caller is responsible for the original address string.
  *
- * ~~~~~~~~~~{.c}
+ * @code
  * as_config config;
  * as_config_init(&config);
  * as_config_tls_add_host(&config, "127.0.0.1", "node1.test.org", 3000);
- * ~~~~~~~~~~
+ * @endcode
  *
  * @relates as_config
  */
 AS_EXTERN void
 as_config_tls_add_host(as_config* config, const char* address, const char* tls_name, uint16_t port);
 
+/**
+ * Set dynamic configuration file path.
+ *
+ * An altenate way to set the path is to set environment variable AEROSPIKE_CLIENT_CONFIG_URL
+ * before running the application.
+ *
+ * @relates as_config
+ */
+static inline void
+as_config_provider_set_path(as_config* config, const char* path)
+{
+	as_config_set_string(&config->config_provider.path, path);
+}
+
+/**
+ * Set dynamic configuration file path and file modification check interval.
+ * The interval is the number of cluster tend iterations.
+ *
+ * @relates as_config
+ */
+static inline void
+as_config_provider_set(as_config* config, const char* path, uint32_t interval)
+{
+	as_config_provider_set_path(config, path);
+	config->config_provider.interval = interval;
+}
+
+/**
+ * Add rack id to list of server racks in order of preference. Only add racks that
+ * are close to the client rack. Do not add racks that are far away from the client.
+ * The client will still direct commands to far away racks if nodes on closer racks are not
+ * available.
+ *
+ * rack_aware, AS_POLICY_REPLICA_PREFER_RACK and server rack configuration must also be
+ * set to enable this functionality.
+ *
+ * @code
+ * as_config config;
+ * as_config_init(&config);
+ * // Rack 4 is where the client machine is located.
+ * as_config_add_rack_id(&config, 4);
+ * // Rack 2 is located in same datacenter complex (maybe in a close by building).
+ * as_config_add_rack_id(&config, 2);
+ * // All other racks are far enough away that they are equally not preferred, so do not include
+ * // them here.
+ * @endcode
+ *
+ * @relates as_config
+ */
+AS_EXTERN void
+as_config_add_rack_id(as_config* config, int rack_id);
+
+/**
+ * Convert string into as_auth_mode enum.
+ */
+AS_EXTERN bool
+as_auth_mode_from_string(as_auth_mode* auth, const char* str);
+
+/**
+ * @private
+ * Validate and modify max_error_rate and error_rate_window if not within bounds.
+ */
+void
+as_config_massage_error_rate(as_config* config);
+
 #ifdef __cplusplus
 } // end extern "C"
 #endif
